import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;

import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;

import com.nimbusds.oauth2.sdk.AuthorizationCode;
import com.nimbusds.oauth2.sdk.AuthorizationCodeGrant;
import com.nimbusds.oauth2.sdk.AuthorizationRequest;
import com.nimbusds.oauth2.sdk.ResponseType;
import com.nimbusds.oauth2.sdk.Scope;
import com.nimbusds.oauth2.sdk.TokenRequest;
import com.nimbusds.oauth2.sdk.TokenResponse;
import com.nimbusds.oauth2.sdk.TokenRevocationRequest;
import com.nimbusds.oauth2.sdk.auth.ClientAuthentication;
import com.nimbusds.oauth2.sdk.auth.ClientSecretBasic;
import com.nimbusds.oauth2.sdk.auth.Secret;
import com.nimbusds.oauth2.sdk.http.HTTPResponse;
import com.nimbusds.oauth2.sdk.id.ClientID;
import com.nimbusds.oauth2.sdk.id.State;
import com.nimbusds.oauth2.sdk.token.AccessToken;
import com.nimbusds.oauth2.sdk.token.BearerAccessToken;
import com.nimbusds.openid.connect.sdk.OIDCTokenResponse;
import com.nimbusds.openid.connect.sdk.UserInfoRequest;
import com.nimbusds.openid.connect.sdk.UserInfoResponse;

import jakarta.servlet.http.HttpServlet;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;

public class OAuth2AuthorizationCodeFlow {

    private static final String CLIENT_ID = "SR_91uGHZXCp00zZcJ8A6";
    private static final String CLIENT_SECRET = "3ZfhJrhmYVMOw51UFzVdKU6F1JeIMm9P8GSx8FpWOdtUJ3KduSxOHVNgOPeez5_fcF-5l3e0lg3c7yJw8WfkIA";
    private static final String REDIRECT_URI = "http://localhost:3000/cb";
    private static final String AUTHORIZATION_ENDPOINT = "https://login.pre.nautilus.app/oidc/authorize";
    private static final String TOKEN_ENDPOINT = "https://login.pre.nautilus.app/oidc/oauth/token";
    private static final String USERINFO_ENDPOINT = "https://login.pre.nautilus.app/oidc/me";
    private static final String REVOCATION_ENDPOINT = "https://login.pre.nautilus.app/oidc/oauth/revoke";
    private static final String END_SESSION_ENDPOINT = "https://login.pre.nautilus.app/oidc/v2/logout";
    private static final String POST_LOGOUT_REDIRECT_URI = "http://localhost:3000/post-logout";
    private static final String SCOPES = "openid email profile";

    public static void main(String[] args) throws Exception {
        try {
            Server server = new Server(3000);

            ServletContextHandler context = new ServletContextHandler(ServletContextHandler.SESSIONS);
            context.setContextPath("/");
            server.setHandler(context);

            context.addServlet(new ServletHolder(new CallbackServlet()), "/cb");

            server.start();

            AuthorizationRequest authRequest = new AuthorizationRequest.Builder(
                    new ResponseType(ResponseType.Value.CODE),
                    new ClientID(CLIENT_ID))
                    .endpointURI(URI.create(AUTHORIZATION_ENDPOINT))
                    .scope(new Scope(SCOPES))
                    .state(new State())
                    .redirectionURI(URI.create(REDIRECT_URI))
                    .build();

            URI authURI = authRequest.toURI();
            System.out.println("Visit the following URL to authorize:");
            System.out.println(authURI.toString());

            System.out.println("Server started at http://localhost:3000/cb");
            server.join();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static class CallbackServlet extends HttpServlet {
        @Override
        protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws IOException {
            String code = req.getParameter("code");
            if (code != null) {
                try {
                    AuthorizationCode authCode = new AuthorizationCode(code);
                    URI tokenEndpoint = URI.create(TOKEN_ENDPOINT);

                    ClientAuthentication clientAuth = new ClientSecretBasic(new ClientID(CLIENT_ID), new Secret(CLIENT_SECRET));

                    TokenRequest tokenRequest = new TokenRequest(
                            tokenEndpoint,
                            clientAuth,
                            new AuthorizationCodeGrant(authCode, URI.create(REDIRECT_URI))
                    );

                    HTTPResponse tokenHTTPResponse = tokenRequest.toHTTPRequest().send();
                    TokenResponse tokenResponse = TokenResponse.parse(tokenHTTPResponse);

                    if (tokenResponse instanceof OIDCTokenResponse) {
                        OIDCTokenResponse successResponse = (OIDCTokenResponse) tokenResponse;
                        AccessToken accessToken = successResponse.getOIDCTokens().getAccessToken();
                        String idToken = successResponse.getOIDCTokens().getIDToken().getParsedString();

                        // Llamada al userinfo_endpoint antes de la revocación
                        System.out.println("Fetching user info before revocation...");
                        fetchUserInfo(accessToken, resp);

                        // Revocar el token de acceso
                        revokeAccessToken(accessToken, resp);

                        // Llamada al userinfo_endpoint después de la revocación
                        System.out.println("Fetching user info after revocation...");
                        fetchUserInfo(accessToken, resp);

                        // Llamar a la función de logout
                        logout(idToken, resp);
                    } else {
                        resp.getWriter().println("Error: " + tokenResponse.toErrorResponse().getErrorObject().getDescription());
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                    resp.getWriter().println("Error processing callback: " + e.getMessage());
                }
            } else {
                resp.getWriter().println("Authorization code not found in callback.");
            }
        }

        private void fetchUserInfo(AccessToken accessToken, HttpServletResponse resp) throws IOException {
            try {
                URI userInfoEndpoint = URI.create(USERINFO_ENDPOINT);
                UserInfoRequest userInfoRequest = new UserInfoRequest(userInfoEndpoint, (BearerAccessToken) accessToken);
                HTTPResponse userInfoHTTPResponse = userInfoRequest.toHTTPRequest().send();
                UserInfoResponse userInfoResponse = UserInfoResponse.parse(userInfoHTTPResponse);

                if (userInfoResponse.indicatesSuccess()) {
                    resp.getWriter().println("User Info: " + userInfoResponse.toSuccessResponse().getUserInfo().toJSONObject());
                } else {
                    resp.getWriter().println("Failed to fetch user info: " + userInfoResponse.toErrorResponse().getErrorObject().getDescription());
                }
            } catch (Exception e) {
                resp.getWriter().println("Error fetching user info: " + e.getMessage());
            }
        }

        private void revokeAccessToken(AccessToken accessToken, HttpServletResponse resp) throws IOException {
            try {
                URI revocationEndpoint = URI.create(REVOCATION_ENDPOINT);
                TokenRevocationRequest revocationRequest = new TokenRevocationRequest(
                        revocationEndpoint,
                        new ClientSecretBasic(new ClientID(CLIENT_ID), new Secret(CLIENT_SECRET)),
                        accessToken
                );

                HTTPResponse revocationHTTPResponse = revocationRequest.toHTTPRequest().send();
                if (revocationHTTPResponse.getStatusCode() == HTTPResponse.SC_OK) {
                    resp.getWriter().println("Access token revoked successfully.");
                } else {
                    resp.getWriter().println("Failed to revoke access token: " + revocationHTTPResponse.getContent());
                }
            } catch (Exception e) {
                resp.getWriter().println("Error revoking access token: " + e.getMessage());
            }
        }

        private void logout(String idToken, HttpServletResponse resp) throws IOException {
            try {
                URI logoutUri = new URI(END_SESSION_ENDPOINT + "?id_token_hint=" + idToken + "&post_logout_redirect_uri=" + POST_LOGOUT_REDIRECT_URI);
                resp.sendRedirect(logoutUri.toString());
            } catch (URISyntaxException e) {
                e.printStackTrace();
                resp.getWriter().println("Error constructing logout URL: " + e.getMessage());
            }
        }
    }
}
